from tonic_textual.classes.datasetfile import DatasetFile
from typing import List
import requests

class DatasetFileService:
    def __init__(self, client):
        self.client = client

    def get_files(self, dataset_id: str) -> List[DatasetFile]:
        with requests.Session() as session:
            dataset = self.client.http_get(f"/api/dataset/{dataset_id}", session=session)
            return [DatasetFile(f['fileId'],f['fileName'],f.get('numRows'),f['numColumns'], f['processingStatus'], f.get('processingError')) for f in dataset["files"]]



import requests
from urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

class HttpClient:
    """Client used to handle requests to the Tonic Textual instance.

    Parameters
    ----------
    base_url : str
        URL to the Tonic Textual instance.
    api_key : str
        The API token associated to use for the requests.
    """

    def __init__(self, base_url: str, api_key: str):
        self.base_url = base_url
        self.headers = {"Authorization": api_key, "User-Agent": "tonic-textual-python-sdk"}

    def http_get_file(self, url: str, session: requests.Session, params: dict={}) -> str:
        """Makes a get request to get a file.

        Parameters
        ----------
        url : str
            URL to make the get request. The URL is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.get request.

        """
        res = session.get(self.base_url + url, params=params, headers=self.headers, verify=False)
        res.raise_for_status()
        return res.content.decode('utf-8')

    def http_get(self, url: str, session: requests.Session, params: dict={}):
        """Makes a get request.

        Parameters
        ----------
        url : str
            URL to make the get request. The URL is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.get request.

        """
        res = session.get(self.base_url + url, params=params, headers=self.headers, verify=False)
        res.raise_for_status()
        return res.json()

    def http_post(self, url, params={}, data={}, files={}):
        """Make a post request.

        Parameters
        ----------
        url : str
            URL to make the post request. The URL is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.post request.
        data: dict
            Passed as the data parameter of the requests.post request.
        """
        res = requests.post(
            self.base_url + url, params=params, json=data, headers=self.headers, verify=False, files=files
        )
        res.raise_for_status()
        return res.json()

    def http_put(self, url, params={}, data={}, files={}):
        """Makes a put request.

        Parameters
        ----------
        url : str
            URL to make the put request. The URL is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.put request.
        data: dict
            Passed as the data parameter of the requests.put request.
        """
        res = requests.put(
            self.base_url + url, params=params, json=data, headers=self.headers, verify=False
        )
        res.raise_for_status()
        return res.json()

    def http_delete(self, url, params={}):
        res = requests.delete(self.base_url + url, params = params, headers = self.headers, verify = False)
        res.raise_for_status()
        return res.json()



class DatasetFile:
    """
    Class to store the metadata for a dataset file.

    Parameters
    ----------
    id : str
        The identifier of the dataset file.

    name: str
        The file name of the dataset file.

    num_rows : long
        The number of rows in the dataset file.

    num_columns: int
        The number of columns in the dataset file.

    processing_status: string
        The status of the dataset file in the processing pipeline. Possible values are 'Completed', 'Failed', 'Cancelled', 'Running', and 'Queued'.
    
    processing_error: string
        If the dataset file processing failed, a description of the issue that caused the failure.

    uploaded_timestamp: str
        Timestamp in UTC when dataset file was uploaded to the dataset.
    """
    def __init__(self, id: str, name: str, num_rows: int, num_columns: int, processing_status: str, processing_error: str):
        self.id = id
        self.name = name
        self.num_rows = num_rows
        self.num_columns = num_columns
        self.processing_status = processing_status
        self.processing_error = processing_error

    def describe(self):
        """Prints the dataset file metadata. Includes the identifier, file name, number of rows, and number of columns.
        """
        print("File: " + self.name + " [" + self.id + "]")
        print("Number of rows: " + self.num_rows)
        print("Number of columns: " + self.num_columns)
        print("Status: " + self.processing_status)
        if self.processing_status!="" and self.processing_error is not None:
            print("Error: " + self.processing_error)



import os
import requests;

from typing import IO, List, Optional
from urllib.parse import urlencode

from tonic_textual.classes.api_responses.single_detection_result import SingleDetectionResult
from tonic_textual.classes.custom_model import CustomModel
from tonic_textual.classes.generator_config import GeneratorConfig
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.api_responses.redaction_response import RedactionResponse
from tonic_textual.enums.pii_state import PiiState
from tonic_textual.services.dataset import DatasetService
from tonic_textual.services.datasetfile import DatasetFileService
from tonic_textual.classes.dataset import Dataset
from tonic_textual.classes.datasetfile import DatasetFile
from tonic_textual.classes.tonic_exception import DatasetNameAlreadyExists, InvalidJsonForRedactionRequest

class TonicTextual:
    '''Wrapper class for invoking Tonic Textual API

    Parameters
    ----------
    base_url : str
        The URL to your Tonic Textual instance. Do not include trailing backslashes.
    api_key : str
        Your API token. This argument is optional. Instead of providing the API token here, it is recommended that you set the API key in your environment as the value of TONIC_TEXTUAL_API_KEY.

    Examples
    --------
    >>> TonicTextual("http://localhost:3000")
    '''
    def __init__(self, base_url : str, api_key: Optional[str] = None):
        if api_key is None:
            api_key = os.environ.get("TONIC_TEXTUAL_API_KEY")
            if api_key is None:
                raise Exception("No API key provided. Either provide an API key, or set the API key as the value of the TONIC_TEXTUAL_API_KEY environment variable.")
        self.api_key = api_key
        self.client = HttpClient(base_url, self.api_key)
        self.dataset_service = DatasetService(self.client)
        self.datasetfile_service = DatasetFileService(self.client)

    def create_dataset(self, dataset_name:str):
        """Creates a dataset. A dataset is a collection of 1 or more files for Tonic Textual to scan and redact.

        Parameters
        -----
        dataset_name : str
            The name of the dataset. Dataset names must be unique.


        Returns
        -------
        Dataset
            The newly created dataset.


        Raises
        ------

        DatasetNameAlreadyExists
            Raised if a dataset with the same name already exists.

        """
        try:
            self.client.http_post("/api/dataset", data={"name": dataset_name})
        except requests.exceptions.HTTPError as e:
            if e.response.status_code==409:
                raise DatasetNameAlreadyExists(e)

        return self.get_dataset(dataset_name)

    def delete_dataset(self, dataset_name: str):
        params = { "datasetName": dataset_name}
        self.client.http_delete("/api/dataset/delete_dataset_by_name?" + urlencode(params))




    def get_dataset(self, dataset_name : str) -> Dataset:
        '''Gets the dataset for the specified dataset name.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset.

        Returns
        -------
        Dataset

        Examples
        --------
        >>> dataset = tonic.get_dataset("llama_2_chatbot_finetune_v5")
        '''
        return self.dataset_service.get_dataset(dataset_name)

    def get_files(self, dataset_id: str) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset.

        Returns
        ------
        List[DatasetFile]
        A list of all of the files in the dataset.
        """
        return self.datasetfile_service.get_files(dataset_id)
      
    def unredact_bulk(self, redacted_strings: List[str]) -> List[str]:
            """Removes redaction from a list of strings. Returns the strings with the original values.
            
            Parameters
            ----------
            redacted_strings : List[str]
                The list of redacted strings from which to remove the redaction.
    
            Returns
            -------
            List[str]
                The list of strings with the redaction removed.
            """
            
            response = self.client.http_post("/api/unredact", data=redacted_strings)            
            return response
    
    def unredact(self, redacted_string: str) -> str:
            """Removes the redaction from a provided string. Returns the string with the original values.
            
            Parameters
            ----------
            redacted_string : str
                The redacted string from which to remove the redaction.
    
            Returns
            -------
            str
                The string with the redaction removed.
            """
            
            response = self.client.http_post("/api/unredact", data=[redacted_string])            
            return response
    
    def redact(self, string: str, generatorConfig: GeneratorConfig = dict(), customModels: List[str] = []) -> RedactionResponse:
            """Redacts a string. Depending on the configured handling for each sensitive data type, values can be either redacted, synthesized, or ignored.
            
            Parameters
            ----------
            string : str
                The string to redact.
            
            generatorConfig: GeneratorConfig
                A dictionary of sensitive data entities. For each entity, indicates whether to redact, synthesize, or ignore it.
    
            customModels: List[str]
                A list of custom model names to use to identify values to redact. To see the list of custom models that you have access to, use the get_custom_models function.

            Returns
            -------
            RedactionResponse
                The redacted string along with ancillary information.
            """
            
            invalid_pii_states = [v for v in list(generatorConfig.values()) if v not in PiiState._member_names_]
            if(len(invalid_pii_states)>0):
                 raise Exception("Invalid configuration for generatorConfig. The allowed values are Off, Synthesis, and Redaction.")                 

            endpoint = "/api/redact"
            response = self.client.http_post(endpoint, data={"text": string, "generatorConfig": generatorConfig, "customModels": customModels})
            
            de_id_results = [SingleDetectionResult(x["start"], x["end"], x["label"], x["text"], x["score"]) for x in list(response["deIdentifyResults"])]

            return RedactionResponse(response["originalText"], response["redactedText"], response["usage"], de_id_results)
    
    def redact_json(self, json_string: str, generatorConfig: GeneratorConfig = dict(), customModels: List[str] = []) -> RedactionResponse:
        """Redacts a string. Depending on the configured handling for each sensitive data type, values can be either redacted, synthesized, or ignored.
        
        Parameters
        ----------
        json_string : str
            The JSON string to redact
        
        generatorConfig: GeneratorConfig
            A dictionary of sensitive data entities. For each entity, indicates whether to redact, synthesize, or ignore it.

        customModels: List[str]
            A list of custom model names to use to identify values to redact. To see the list of custom models that you have access to, use the get_custom_models function.

        Returns
        -------
        RedactionResponse
            The redacted string along with ancillary information.
        """
        
        invalid_pii_states = [v for v in list(generatorConfig.values()) if v not in PiiState._member_names_]
        if(len(invalid_pii_states)>0):
                raise Exception("Invalid configuration for generatorConfig. The allowed values are Off, Synthesis, and Redaction.")                 

        endpoint = "/api/redact/json"
        try:
            response = self.client.http_post(endpoint, data={"jsonText": json_string, "generatorConfig": generatorConfig, "customModels": customModels})
        except requests.exceptions.HTTPError as e:
             if e.response.status_code==400:                  
                  raise InvalidJsonForRedactionRequest(e.response.text)
             raise e
        
        de_id_results = [SingleDetectionResult(x["start"], x["end"], x["label"], x["text"], x["score"], x["jsonPath"]) for x in list(response["deIdentifyResults"])]

        return RedactionResponse(response["originalText"], response["redactedText"], response["usage"], de_id_results)


    def get_custom_models(self) -> List[CustomModel]:
        """Returns all of the custom models that the user owns.
         
        Returns
        -------
        List[CustomModel]
            A list of all of the custom models that the user owns.
        """

        with requests.Session() as session:
            response = self.client.http_get("/api/models", session=session)
            models: List[CustomModel] = []
            for model in response:
                id = model['id']
                name = model['name']
                entities = model['entities']
                entityNames = [entity['label'] for entity in entities]
                models.append(CustomModel(id, name, entityNames))
            
            return models



from typing import List

from tonic_textual.classes.api_responses.single_detection_result import SingleDetectionResult

class RedactionResponse(dict):
    def __init__(self, original_text: str, redacted_text: str, usage: int, de_identify_results: List[SingleDetectionResult]):
        self.original_text = original_text
        self.redacted_text = redacted_text
        self.usage = usage
        self.de_identify_results = de_identify_results
        dict.__init__(self, original_text=original_text, redacted_text=redacted_text, usage=usage, de_identify_results=de_identify_results)

    def describe(self):
        print(self.redacted_text)        
        print('')
        for x in self.de_identify_results:
            x.describe()

    def get_usage(self):
        return self.usage


from requests.exceptions import HTTPError

class DatasetNameAlreadyExists(Exception):
    """
        Raised when there is an attempt to create a dataset with a name that already exists.
    """
    def __init__(self, errors):
        # Call the base class constructor with the parameters it needs
        super().__init__("The dataset name already exists. Dataset names must be unique. Choose a different name.")
        self.errors = errors

class DatasetFileMatchesExistingFile(HTTPError):
    """
        Raised when the content in a file to upload matches the content in an existing file in the dataset.
    """
    def __init__(self, errors):
        super().__init__(errors.response.content or "The file content matches content in an existing dataset file. Choose a different file.")
        self.errors = errors

class InvalidJsonForRedactionRequest(Exception):
    """
        Raised when the JSON redaction request contains invalid JSON
    """
    def __init__(self, msg):
        # Call the base class constructor with the parameters it needs
        super().__init__(msg)


from typing import List

class CustomModel:
    def __init__(self, id: str, name: str, entities: List[str]):
        self.id = id
        self.name = name
        self.entities = entities

    def describe(self):
        print("Model Name: " + self.name)
        print("Model Id: " + self.id)
        print("Entities: " + ', '.join(self.entities))



import json
from typing import Optional

class SingleDetectionResult(dict):
    def __init__(self, start: int, end: int, label: str, text: str, score: float, json_path: Optional[str]):
        self.start = start
        self.end = end
        self.label = label
        self.text = text
        self.score = score
        self.jsonPath = json_path
        if json_path is None:
            dict.__init__(self, start=start, end=end, label=label, text=text, score=score)
        else:
            dict.__init__(self, start=start, end=end, label=label, text=text, score=score, json_path=json_path)
            
    def describe(self):
        print(json.dumps(self))


from __future__ import annotations
from typing import List, Dict, Optional
import os
import json
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper
from urllib.parse import urlencode
import requests.exceptions;
import requests
import pandas as pd
from tonic_textual.classes.tonic_exception import DatasetFileMatchesExistingFile
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.datasetfile import DatasetFile
from tonic_textual.services.datasetfile import DatasetFileService


class Dataset:
    """
    Class to represent and provide access to a Tonic Textual dataset.

    Parameters
    ----------
    id: str
        Dataset identifier.

    name: str
        Dataset name.

    files: Dict
        Serialized DatasetFile objects representing the files in a dataset.

    client: HttpClient
        The HTTP client to use.
    """
    def __init__(self, id: str, name: str, files: List[Dict], client: HttpClient):
        self.id = id
        self.name = name
        self.client = client
        self.datasetfile_service = DatasetFileService(self.client)
        self.files = [DatasetFile(f['fileId'],f['fileName'],f.get('numRows'),f['numColumns'], f['processingStatus'], f.get('processingError')) for f in files]

        if len(self.files) > 0:
            self.num_columns = max([f.num_columns for f in self.files])
        else:
            self.num_columns = None

        self.num_rows_per_request = 1000
        self.total_rows_fetched = 0

        self._total_rows_fetched_in_current_file = 0
        self._cur_file_idx = 0

    def reset(self):
        self.total_rows_fetched = 0
        self._cur_file_idx = 0
        self._total_rows_fetched_in_current_file = 0

    def upload_then_add_file(self, file_path: str, file_name: Optional[str] = None):
        """
        Uploads a file to the dataset.

        Parameters
        --------
        file_path: str
            The absolute path of the file to upload.
        file_name: str
            The name of the file to save to Tonic Textual.

        Raises
        ------

        DatasetFileMatchesExistingFile
            Returned if the file content matches an existing file.

        """
        if file_name is None:
            file_name = os.path.basename(file_path)
        file_size = os.path.getsize(file_path)
        with open(file_path, 'rb') as f:
            with tqdm(desc=f"[INFO] Uploading", total=file_size, unit="B", unit_scale=True, unit_divisor=1024) as t:
                reader_wrapper = CallbackIOWrapper(t.update, f, "read")

                files = {
                    'document': (None, json.dumps({"fileName": file_name, "csvConfig":{}, "datasetId": self.id}), 'application/json'),
                    'file': reader_wrapper
                }
                try:
                    updated_dataset = self.client.http_post(f"/api/dataset/{self.id}/files/upload", files=files)
                    #numRows is null when a file is first uploaded
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code==409:
                        raise DatasetFileMatchesExistingFile(e)
                    else:
                        raise e
        self.files = [DatasetFile(f['fileId'],f['fileName'],f.get('numRows'),f['numColumns'], f['processingStatus'], f.get('processingError')) for f in updated_dataset["files"]]
        self.num_columns = max([f.num_columns for f in self.files])

    def fetch_all_df(self) -> pd.DataFrame:
        """
        Fetches all of the data in the dataset as a pandas dataframe.

        Returns
        -------
        pd.DataFrame
            Dataset data in a pandas dataframe.
        """
        data = self._fetch_all()

        if self.num_columns is None:
            return pd.DataFrame()

        #RAW file, not CSV
        if self.num_columns == 0:
            if len(data)==0:
                return pd.DataFrame(columns=["text"])
            return pd.DataFrame(data, columns=["text"])

        columns = ['col'+str(x) for x in range(self.num_columns)]
        if len(data)==0:
            return pd.DataFrame(columns=columns)
        else:
            return pd.DataFrame(data, columns=columns)

    def fetch_all_json(self) -> str:
        """
        Fetches all of the data in the dataset as JSON.

        Returns
        -------
        str
            Dataset data in JSON format.
        """
        return json.dumps(self._fetch_all())

    def _fetch_all(self) -> List[List[str]]:
        """
        Fetches all data from the dataset.

        Returns
        -------
        List[List[str]]
            The datset data.
        """
        response = []
        with requests.Session() as session:
            for file in self.files:
                try:
                    if file.num_columns==0:
                        more_data = self.client.http_get_file(f"/api/dataset/{self.id}/files/{file.id}/get_data", session=session)
                        response+=[[more_data]]
                    else:
                        more_data = self.client.http_get(f"/api/dataset/{self.id}/files/{file.id}/get_data", session=session)
                        response += more_data
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code==409:
                        continue
                    else:
                        raise e
            return response

    def get_processed_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset for which processing is complete. The data in these files is returned when data is requested.

        Returns
        ------
        List[DatasetFile]:
            The list of processed dataset files.
        """
        return list(filter(lambda x: x.processing_status=="Completed" ,self.files))

    def get_queued_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset that are waiting to be processed.

        Returns
        ------
        List[DatasetFile]:
            The list of dataset files that await processing.
        """
        return list(filter(lambda x: x.processing_status=="Queued" ,self.files))

    def get_running_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset that are currently being processed.

        Returns
        ------
        List[DatasetFile]:
            The list of files that are being processed.
        """
        return list(filter(lambda x: x.processing_status=="Running" ,self.files))

    def get_failed_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in dataset that encountered an error when they were processed. These files are effectively ignored.

        Returns
        ------
        List[DatasetFile]:
            The list of files that had processing errors.
        """
        return list(filter(lambda x: x.processing_status=="Failed" ,self.files))

    def _check_processing_and_update(self):
        """
        Checks the processing status of the files in the dataset and updates the files list.
        """
        if len(self.get_queued_files()+self.get_running_files()) > 0:
            self.files = self.datasetfile_service.get_files(self.id)

    def describe(self):
        """
        Prints the dataset name, identifier, and the list of files.

        Examples
        --------
        >>> workspace.describe()
        Dataset: your_dataset_name [dataset_id]
        Number of Files: 2
        Number of Rows: 1000
        """
        self._check_processing_and_update()

        print("Dataset: " + self.name + " [" + self.id + "]")
        print("Number of Files: " + str(len(self.get_processed_files())))
        print("Files that are waiting for processing: " + ", ".join([str((f.id,f.name)) for f in self.get_queued_files()+self.get_running_files()]))
        print("Files that encountered errors while processing: " + ", ".join([str((f.id,f.name)) for f in self.get_failed_files()]))
        print("Number of Rows: " + str(sum([x.num_rows if x.num_rows is not None else 0 for x in self.files])))
        print("Number of rows fetched: " + str(self.total_rows_fetched))
